"""
This module implements incremental reduced error pruning (IREP*) algorithm
for growing classification rulesets.
See https://www.let.rug.nl/nerbonne/teach/learning/cohen95fast.pdf
"""

# Author: Ilan Moscovitz <ilan.moscovitz@gmail.com>
# License: MIT

import pandas as pd
import math
import random
import copy
import warnings
import numpy as np

from wittgenstein import base
from .base import Cond, Rule, Ruleset
from .base import rnd, score_accuracy, bin_df

class IREP:
    """ Class for generating ruleset classification models. """

    def __init__(self, prune_size=.33, verbosity=0):
        """ Creates a new IREP object.

            args:
                prune_size (optional):   proportion of training set to be used for pruning (defailt=.33).
                                           Set to None to skip pruning (not recommended).
                verbosity (optional):    output information about the training process (default=0)
                                           1: Show results of each major phase
                                           2: Show Ruleset grow/optimization steps
                                           3: Show Ruleset grow/optimization calculations
                                           4: Show Rule grow/prune steps
                                           5: Show Rule grow/prune calculations
        """
        self.prune_size = prune_size
        self.verbosity = verbosity

    def __str__(self):
        """ Returns string representation of an IREP object. """
        fitstr = f'with fit ruleset' if hasattr(self,'ruleset_') else '(unfit)'
        return f'<IREP object {fitstr}>'
    __repr__ = __str__

    def out_model(self):
        """ Prints trained Ruleset model line-by-line: V represents 'or'; ^ represents 'and'. """
        if hasattr(self,'ruleset_'):
            self.ruleset_.out_pretty()
        else:
            print('no model fitted')

    def fit(self, df, y=None, class_feat=None, pos_class=None, n_discretize_bins=10, max_rules=None, max_rule_conds=None, max_total_conds=None, random_state=None):
        """ Fit a Ruleset model using a training DataFrame.

            args:
                df <DataFrame>: categorical training dataset
                y: <iterable>: class labels corresponding to df rows. Parameter y or class_feat (see next) must be provided.
                class_feat: column name of class feature (Use if class feature is still in df.)

                pos_class (optional): name of positive class. If not provided, defaults to class of first training example.
                n_discretize_bins (optional): Fit apparent numeric attributes into a maximum of n_discretize_bins discrete bins, inclusive on upper part of range.
                                              Pass None to disable auto-discretization. (default=10)

                random_state: (optional) random state to allow for repeatable results

                options to stop early. Intended for improving model interpretability or limiting training time on noisy datasets. Not specifically intended for use as a hyperparameter, since pruning already occurs during training, though it is certainly possible that tuning could improve model performance.
                max_rules (optional): maximum number of rules. default=None
                max_rule_conds (optional): maximum number of conds per rule. default=None
                max_total_conds (optional): maximum number of total conds in entire ruleset. default=None

        """

        # Stage 0: Setup

        # Set up trainset, set class feature name, and set pos class name
        df, self.class_feat, self.pos_class = base.trainset_classfeat_posclass(df, y=y, class_feat=class_feat, pos_class=pos_class)

        # Anything to discretize?
        df, self.bin_transformer_ = bin_df(df, n_discretize_bins=n_discretize_bins, ignore_feats=[self.class_feat], verbosity=self.verbosity)

        # Split df into pos, neg classes
        pos_df, neg_df = base.pos_neg_split(df, self.class_feat, self.pos_class)
        pos_df = pos_df.drop(self.class_feat,axis=1)
        neg_df = neg_df.drop(self.class_feat,axis=1)

        # Stage 1 (of 1): Grow Ruleset
        self.ruleset_ = Ruleset()
        self.ruleset_ = self._grow_ruleset(pos_df, neg_df,
            prune_size=self.prune_size, max_rules=max_rules, max_rule_conds=max_rule_conds, max_total_conds=max_total_conds, random_state=random_state)
        if self.verbosity >= 1:
            print()
            print('GREW RULESET:')
            self.ruleset_.out_pretty()
            print()

        # Fit probas
        self.recalibrate_proba(df, min_samples=None, require_min_samples=False, discretize=False)

    def predict(self, X_df, give_reasons=False):
        """ Predict classes of data using a IREP-fit model.

            args:
                X_df <DataFrame>: examples to make predictions on.

                give_reasons (optional) <bool>: whether to provide reasons for each prediction made.

            returns:
                list of <bool> values corresponding to examples. True indicates positive predicted class; False non-positive class.

                If give_reasons is True, returns a tuple that contains the above list of predictions
                    and a list of the corresponding reasons for each prediction;
                    for each positive prediction, gives a list of all the covering Rules, for negative predictions, an empty list.
        """

        if not hasattr(self, 'ruleset_'):
            raise AttributeError('You should fit an IREP object before making predictions with it.')

        if not self.bin_transformer_:
            return self.ruleset_.predict(X_df, give_reasons=give_reasons)
        else:
            binned_X = X_df.copy()
            base.bin_transform(binned_X, self.bin_transformer_)
            return self.ruleset_.predict(binned_X, give_reasons=give_reasons)

    def score(self, X, y, score_function=score_accuracy):
        """ Test performance of an IREP-fit model.

            X: <DataFrame> of independent attributes
            y: <DataFrame> or <iterable> of matching dependent target values

            score_function (optional): function that takes two parameters: actuals <iterable<bool>>, predictions <iterable<bool>>,
                                       containing class values. (default=accuracy)
                                       this parameter is intended to be compatible with sklearn's scoring functions:
                                       https://scikit-learn.org/stable/modules/model_evaluation.html#classification-metrics
        """

        predictions = self.predict(X)
        if type(y)==pd.core.frame.DataFrame:
            actuals = [yi==self.pos_class for yi in y.tolist()]
        else:
            actuals = [yi==self.pos_class for yi in y]
        return score_function(actuals, predictions)

    def predict_proba(self, X_df, give_reasons=False, ret_n=False, min_samples=1, discretize=True):
        # Drop class feature if user forgot to:
        df = X_df if self.class_feat not in X_df.columns else X_df.drop(self.class_feat, axis=1)
        return self.ruleset_.predict_proba(df, give_reasons=give_reasons, ret_n=ret_n, min_samples=min_samples, discretize=True, bin_transformer=self.bin_transformer_)

    def recalibrate_proba(self, Xy_df, min_samples=20, require_min_samples=True, discretize=True):
        """ Recalibrate a classifier's probability estimations using unseen labeled data. May improve .predict_proba generalizability.
            Does not affect the underlying model or which predictions it makes -- only probability estimates. Use params min_samples and require_min_samples to select desired behavior.

            Note1: RunTimeWarning will occur as a reminder when min_samples and require_min_samples params might result in unintended effects.
            Note2: It is possible recalibrating could result in some positive .predict predictions with <0.5 .predict_proba positive probability.

            Xy_df <DataFrame>: labeled data

            min_samples <int> (optional): required minimum number of samples per Rule
                                          default=10. set None to ignore min sampling requirement so long as at least one sample exists.
            require_min_samples <bool> (optional): True: halt (with warning) in case min_samples not achieved for all Rules
                                                   False: warn, but still replace Rules that have enough samples
            discretize <bool> (optional): If the classifier has already fit a discretization, automatically discretize recalibrate_proba's training data
                                          default=True
        """

        # Recalibrate
        self.ruleset_.recalibrate_proba(Xy_df, class_feat=self.class_feat, pos_class=self.pos_class, min_samples=min_samples, require_min_samples=require_min_samples, discretize=discretize, bin_transformer=self.bin_transformer_)

    def _grow_ruleset(self, pos_df, neg_df, prune_size, max_rules=None, max_rule_conds=None, max_total_conds=None, random_state=None, verbosity=0):
        """ Grow a Ruleset with (optional) pruning. """

        ruleset = Ruleset()
        ruleset._set_possible_conds(pos_df, neg_df)

        if not prune_size: prune_size = 0 # If not pruning, use all the data for growing
        pos_remaining = pos_df.copy()
        neg_remaining = neg_df.copy()
        self.rules = []

        # Stop adding disjunctions if there are no more positive examples to cover
        while (len(pos_remaining) > 0):

            # If applicable, check for user-specified early stopping
            if (max_rules is not None and len(ruleset.rules) >= max_rules) or (max_total_conds is not None and ruleset.count_conds() >= max_total_conds):
                break

            # Grow-prune split remaining uncovered examples (if applicable)
            pos_growset, pos_pruneset = base.df_shuffled_split(pos_remaining, (1-prune_size), random_state=random_state)
            neg_growset, neg_pruneset = base.df_shuffled_split(neg_remaining, (1-prune_size), random_state=random_state)
            if self.verbosity>=2:
                print(f'pos_growset {len(pos_growset)} pos_pruneset {len(pos_pruneset)}')
                print(f'neg_growset {len(neg_growset)} neg_pruneset {len(neg_pruneset)}')
                if not prune_size: print(f'(pruning is turned off)')

            # Grow Rule
            grown_rule = base.grow_rule(pos_growset, neg_growset, ruleset.possible_conds, max_rule_conds=max_rule_conds, verbosity=self.verbosity)

            # If not pruning, add Rule to Ruleset and drop only the covered positive examples
            if not prune_size:
                ruleset.add(grown_rule)
                if self.verbosity>=2:
                    print(f"updated ruleset: {ruleset.truncstr(direction='right')}")
                    print()
                rule_covers_pos = grown_rule.covers(pos_remaining)
                pos_remaining = pos_remaining.drop(rule_covers_pos.index, axis=0)
                if self.verbosity>=3:
                    print(f'examples remaining: {len(pos_remaining)} pos, {len(neg_remaining)} neg')
                    print()

            # If pruning, prune Rule, assess if it's time to stop, and drop all covered examples
            else:
                pruned_rule = base.prune_rule(grown_rule, _IREP_prune_metric, pos_pruneset, neg_pruneset, verbosity=self.verbosity)

                # Stop if the Rule is bad
                prune_precision = base.precision(pruned_rule, pos_pruneset, neg_pruneset)
                if not prune_precision or prune_precision < .50:
                    break
                # Otherwise, add new Rule, remove covered examples, and continue
                else:
                    ruleset.add(pruned_rule)
                    if self.verbosity>=2:
                        print(f"updated ruleset: {ruleset.truncstr(direction='right')}")
                        print()
                    pos_remaining, neg_remaining = base.rm_covered(pruned_rule, pos_remaining, neg_remaining)
                    if self.verbosity>=3:
                        print(f'examples remaining: {len(pos_remaining)} pos, {len(neg_remaining)} neg')
                        print()

        # If applicable, trim total conds
        ruleset.trim_conds(max_total_conds=max_total_conds)

        # Return new ruleset
        return ruleset

##### Metric #####

def _IREP_prune_metric(self, pos_pruneset, neg_pruneset):
    """ Returns the prune value of a candidate Rule """

    P = len(pos_pruneset)
    N = len(neg_pruneset)
    p = self.num_covered(pos_pruneset)
    n = self.num_covered(neg_pruneset)
    return (p+(N - n)) / (P + N)



"""
Implementation of the RIPPERk algorithm for growing classification rulesets.
See https://www.let.rug.nl/nerbonne/teach/learning/cohen95fast.pdf
"""

# Author: Ilan Moscovitz <ilan.moscovitz@gmail.com>
# License: MIT

import pandas as pd
import copy
import math
import warnings
import numpy as np

from wittgenstein import base
from .base import Cond, Rule, Ruleset
from .base import rnd, score_accuracy, bin_df

class RIPPER:
    """ Class for generating ruleset classification models.
        See Cohen (1995): https://www.let.rug.nl/nerbonne/teach/learning/cohen95fast.pdf
    """

    def __init__(self, k=2, prune_size=.33, dl_allowance=64, verbosity=0):
        """ Creates a new RIPPER object.

            args:
                k (optional):            number of RIPPERk optimization iterations (default=2)
                prune_size (optional):   proportion of training set to be used for pruning (default=.33)
                dl_allowance (optional): terminate Ruleset grow phase early if a Ruleset description length is encountered
                                            that is more than this amount above the lowest description length so far encountered.
                                            (default=64 bits)
                verbosity (optional):    output information about the training process (default=0)
                                           1: Show results of each major phase
                                           2: Show Ruleset grow/optimization steps
                                           3: Show Ruleset grow/optimization calculations
                                           4: Show Rule grow/prune steps
                                           5: Show Rule grow/prune calculations
        """
        self.prune_size = prune_size
        self.dl_allowance = dl_allowance
        self.k = k
        self.verbosity = verbosity

    def __str__(self):
        """ Returns string representation of a RIPPER object. """
        fitstr = f'with fit ruleset' if hasattr(self,'ruleset_') else '(unfit)'
        return f'<RIPPER object {fitstr} (k={self.k}, prune_size={self.prune_size}, dl_allowance={self.dl_allowance})>'
    __repr__ = __str__

    def out_model(self):
        """ Prints trained Ruleset model line-by-line: V represents 'or'; ^ represents 'and'. """
        if hasattr(self,'ruleset_'):
            self.ruleset_.out_pretty()
        else:
            print('no model fitted')

    def fit(self, df, y=None, class_feat=None, pos_class=None, n_discretize_bins=10, max_rules=None, max_rule_conds=None, max_total_conds=None, random_state=None):
        """ Fit a Ruleset model using a training DataFrame.

            args:
                df <DataFrame>: categorical training dataset
                y: <iterable>: class labels corresponding to df rows. Parameter y or class_feat (see next) must be provided.
                class_feat: column name of class feature (Use if class feature is still in df.)

                pos_class (optional): name of positive class. If not provided, defaults to class of first training example.
                n_discretize_bins (optional): Fit apparent numeric attributes into a maximum of n_discretize_bins discrete bins, inclusive on upper part of range.
                                              Setting to smaller values can improve training speed.
                                              Pass None to disable auto-discretization and treat values as categorical. (default=10)
                random_state: (optional) random state to allow for repeatable results

                options to stop early. Intended for improving model interpretability or limiting training time on noisy datasets. Not specifically intended for use as a hyperparameter, since pruning already occurs during training, though it is certainly possible that tuning could improve model performance.
                Note: small max_rule_conds may result in the k-optimizations stage making fewer changes.
                max_rules (optional): maximum number of rules. default=None
                max_rule_conds (optional): maximum number of conds per rule. default=None
                max_total_conds (optional): maximum number of total conds in entire ruleset. default=None

        """

        ################
        # Stage 0: Setup
        ################

        # Set up trainset, set class feature name, and set pos class name
        df, self.class_feat, self.pos_class = base.trainset_classfeat_posclass(df, y=y, class_feat=class_feat, pos_class=pos_class)

        # Precalculate rule df lookup
        #self._set_theory_dl_lookup(df, verbosity=self.verbosity)

        # Anything to discretize?
        df, self.bin_transformer_ = bin_df(df, n_discretize_bins=n_discretize_bins, ignore_feats=[self.class_feat], verbosity=self.verbosity)

        # Type as categorical -- this doesn't seem to do anything except slow it down
        #new_type = 'category'
        #print('as type', new_type)
        #df = base.as_type(df, new_type)
        #print(df.dtypes)


        # Split df into pos, neg classes
        pos_df, neg_df = base.pos_neg_split(df, self.class_feat, self.pos_class)
        pos_df = pos_df.drop(self.class_feat,axis=1)
        neg_df = neg_df.drop(self.class_feat,axis=1)

        # Collect possible conds
        self._set_possible_conds(df)

        ###############################
        # Stage 1: Grow initial Ruleset
        ###############################

        self.ruleset_ = Ruleset()
        self.ruleset_ = self._grow_ruleset(pos_df, neg_df,
            prune_size=self.prune_size, dl_allowance=self.dl_allowance, max_rules=max_rules, max_rule_conds=max_rule_conds, max_total_conds=max_total_conds,
            random_state=random_state)
        if self.verbosity >= 1:
            print()
            print('GREW INITIAL RULESET:')
            self.ruleset_.out_pretty()
            print()

        ###########################
        # Stage 2: Optimize Ruleset
        ###########################

        for iter in range(1, self.k+1):
            # Create new but reproducible random_state (if applicable)
            iter_random_state = random_state+100 if random_state is not None else None
            # Run optimization iteration
            if self.verbosity>=1: print(f'optimization run {iter} of {self.k}')
            newset = self._optimize_ruleset(self.ruleset_, pos_df, neg_df, prune_size=self.prune_size, random_state=iter_random_state)

            if self.verbosity>=1:
                print()
                print('OPTIMIZED RULESET:')
                if self.verbosity>=2: print(f'iteration {iter} of {self.k}\n modified rules {[i for i in range(len(self.ruleset_.rules)) if self.ruleset_.rules[i]!= newset.rules[i]]}')
                newset.out_pretty()
                print()

            if iter!= self.k and self.ruleset_ == newset:
                if self.verbosity>=1: print('No changes were made. Halting optimization.')
                break
            else:
                self.ruleset_ = newset

        #############################################
        # Stage 3: Cover any last remaining positives
        #############################################

        pos_remaining, neg_remaining = base.pos_neg_split(df, self.class_feat, self.pos_class)
        pos_remaining = pos_remaining.drop(self.class_feat,axis=1)
        neg_remaining = neg_remaining.drop(self.class_feat,axis=1)
        pos_remaining, neg_remaining = base.rm_covered(self.ruleset_, pos_remaining, neg_remaining)
        if len(pos_remaining)>=1:
            if self.verbosity>=2:
                print(f'{len(pos_remaining)} pos left. Growing final rules...')
            newset = self._grow_ruleset(pos_remaining, neg_remaining, initial_ruleset=self.ruleset_,
                prune_size=self.prune_size, dl_allowance=self.dl_allowance, max_rules=max_rules, max_rule_conds=max_rule_conds, max_total_conds=max_total_conds,
                random_state=random_state)
            if self.verbosity>=1:
                print('GREW FINAL RULES')
                newset.out_pretty()
                print()
            self.ruleset_ = newset
        else:
            if self.verbosity>=1: print('All pos covered\n')

        #################################################
        # Stage 4: Remove any rules that don't improve dl
        #################################################

        if self.verbosity>=2: print('Optimizing dl...')
        mdl_subset, _ = _rs_total_bits(self.ruleset_, self.ruleset_.possible_conds, pos_df, neg_df,
                                        bestsubset_dl=True, ret_bestsubset=True, verbosity=self.verbosity)
        self.ruleset_ = mdl_subset
        if self.verbosity>=1:
            print('FINAL RULESET:')
            self.ruleset_.out_pretty()
            print()

        # Fit probas
        self.recalibrate_proba(df, min_samples=None, require_min_samples=False, discretize=False)

    def predict(self, X_df, give_reasons=False):
        """ Predict classes of data using a RIPPER-fit model.

            args:
                X_df <DataFrame>: examples to make predictions on.

                give_reasons (optional) <bool>: whether to provide reasons for each prediction made.

            returns:
                list of <bool> values corresponding to examples. True indicates positive predicted class; False non-positive class.

                If give_reasons is True, returns a tuple that contains the above list of predictions
                    and a list of the corresponding reasons for each prediction;
                    for each positive prediction, gives a list of all the covering Rules, for negative predictions, an empty list.
        """

        if not hasattr(self, 'ruleset_'):
            raise AttributeError('You should fit a RIPPER object before making predictions with it.')

        if not self.bin_transformer_:
            return self.ruleset_.predict(X_df, give_reasons=give_reasons)
        else:
            binned_X = X_df.copy()
            base.bin_transform(binned_X, self.bin_transformer_)
            return self.ruleset_.predict(binned_X, give_reasons=give_reasons)

    def score(self, X, y, score_function=score_accuracy):
        """ Test performance of a RIPPER-fit model.

            X: <DataFrame> of independent attributes
            y: <DataFrame> or <iterable> of matching dependent target values

            score_function (optional): function that takes two parameters: actuals <iterable<bool>>, predictions <iterable<bool>>,
                                       containing class values. (default=accuracy)
                                       this parameter is intended to be compatible with sklearn's scoring functions:
                                       https://scikit-learn.org/stable/modules/model_evaluation.html#classification-metrics
        """

        predictions = self.predict(X)
        if type(y)==pd.core.frame.DataFrame:
            actuals = [yi==self.pos_class for yi in y.tolist()]
        else:
            actuals = [yi==self.pos_class for yi in y]
        return score_function(actuals, predictions)

    def predict_proba(self, X_df, give_reasons=False, ret_n=False, min_samples=1):
        # Drop class feature if user forgot to:
        df = X_df if self.class_feat not in X_df.columns else X_df.drop(self.class_feat, axis=1)
        return self.ruleset_.predict_proba(df, give_reasons=give_reasons, ret_n=ret_n, min_samples=min_samples, discretize=True, bin_transformer=self.bin_transformer_)

    def recalibrate_proba(self, Xy_df, min_samples=20, require_min_samples=True, discretize=True):
        """ Recalibrate a classifier's probability estimations using unseen labeled data. May improve .predict_proba generalizability.
            Does not affect the underlying model or which predictions it makes -- only probability estimates. Use params min_samples and require_min_samples to select desired behavior.

            Note1: RunTimeWarning will occur as a reminder when min_samples and require_min_samples params might result in unintended effects.
            Note2: It is possible recalibrating could result in some positive .predict predictions with <0.5 .predict_proba positive probability.

            Xy_df <DataFrame>: labeled data

            min_samples <int> (optional): required minimum number of samples per Rule
                                          default=10. set None to ignore min sampling requirement so long as at least one sample exists.
            require_min_samples <bool> (optional): True: halt (with warning) in case min_samples not achieved for all Rules
                                                   False: warn, but still replace Rules that have enough samples
            discretize <bool> (optional): if the classifier has already fit a discretization, automatically discretize recalibrate_proba's training data
                                          default=True
        """

        # Recalibrate
        self.ruleset_.recalibrate_proba(Xy_df, class_feat=self.class_feat, pos_class=self.pos_class, min_samples=min_samples, require_min_samples=require_min_samples, discretize=discretize, bin_transformer=self.bin_transformer_)

    def _set_theory_dl_lookup(self, df, size=15, verbosity=0):
        """ Precalculate rule theory dls for various-sized rules. """

        self.dl_dict = {}

        temp = Ruleset()
        temp._set_possible_conds(df, df)
        possible_conds = temp.possible_conds

        for n in range(1, size+1):
            rule = Rule([Cond('_','_')]*n)
            dl = _r_theory_bits(rule, possible_conds, bits_dict=None, verbosity=verbosity)
            self.dl_dict[n] = dl
            if verbosity>=2:
                print(f'updated dl for rule size {n}: {dl}')

    def _grow_ruleset(self, pos_df, neg_df, prune_size, dl_allowance, initial_ruleset=None, max_rules=None, max_rule_conds=None, max_total_conds=None, random_state=None):
        """ Grow a Ruleset with pruning. """
        pos_remaining = pos_df.copy()
        neg_remaining = neg_df.copy()

        if initial_ruleset is None:
            ruleset = Ruleset()
            ruleset._set_possible_conds(pos_df, neg_df)
        else:
            ruleset = copy.deepcopy(initial_ruleset)

        ruleset_dl = None
        mdl = None      # Minimum encountered description length (in bits)
        dl_diff = 0
        if self.verbosity>=2:
            print('growing ruleset...')
            print()
        while len(pos_remaining) > 0 and dl_diff <= self.dl_allowance:

            # If applicable, check for user-specified early stopping
            if (max_rules is not None and len(ruleset.rules) >= max_rules) or (max_total_conds is not None and ruleset.count_conds() >= max_total_conds):
                break

            # Grow-prune split remaining uncovered examples
            pos_growset, pos_pruneset = base.df_shuffled_split(pos_remaining, (1-prune_size), random_state=random_state)
            neg_growset, neg_pruneset = base.df_shuffled_split(neg_remaining, (1-prune_size), random_state=random_state)
            if self.verbosity>=2:
                print(f'pos_growset {len(pos_growset)} pos_pruneset {len(pos_pruneset)}')
                print(f'neg_growset {len(neg_growset)} neg_pruneset {len(neg_pruneset)}')
            if len(pos_growset)==0: break # Probably safe, but a little dicey to only check pos_growset.

            # Grow Rule
            grown_rule = base.grow_rule(pos_growset, neg_growset, ruleset.possible_conds, max_rule_conds=max_rule_conds, verbosity=self.verbosity)
            if grown_rule.isempty(): break # Generated an empty rule b/c no good conds exist

            # Prune Rule
            pruned_rule = base.prune_rule(grown_rule, _RIPPER_growphase_prune_metric, pos_pruneset, neg_pruneset, verbosity=self.verbosity)

            # Add rule; calculate new description length
            ruleset.add(pruned_rule) # Unlike IREP, IREP*/RIPPER stopping condition is inclusive: "After each rule is added, the total description length of the rule set and examples is computed."
            if self.verbosity>=2:
                print(f"updated ruleset: {ruleset.truncstr(direction='right')}")
                print()

            if ruleset_dl is None:   # First Rule to be added
                rule_dl = _r_theory_bits(pruned_rule, ruleset.possible_conds, verbosity=self.verbosity)
                theory_dl = rule_dl
                data_dl = _exceptions_bits(ruleset, pos_df, neg_df, verbosity=self.verbosity)
                ruleset_dl = theory_dl + data_dl
                mdl = ruleset_dl
            else:
                rule_dl = _r_theory_bits(pruned_rule, ruleset.possible_conds, verbosity=self.verbosity)
                theory_dl += rule_dl
                data_dl = _exceptions_bits(ruleset, pos_df, neg_df, verbosity=self.verbosity)
                ruleset_dl = theory_dl + data_dl
                dl_diff = ruleset_dl - mdl

            if self.verbosity>=3:
                print(f'rule dl: {rnd(rule_dl)}')
                print(f'updated theory dl: {rnd(theory_dl)}')
                print(f'exceptions: {rnd(data_dl)}')
                print(f'total dl: {rnd(ruleset_dl)}')
                if dl_diff<=self.dl_allowance:
                    print(f'mdl {rnd(mdl)} (diff {rnd(dl_diff)} <= {rnd(self.dl_allowance)})')
                else:
                    print(f'mdl {rnd(mdl)} dl-halt: diff {rnd(dl_diff)} exceeds allowance ({rnd(self.dl_allowance)})')

            mdl = ruleset_dl if ruleset_dl<mdl else mdl

            # Remove covered examples
            pos_remaining, neg_remaining = base.rm_covered(pruned_rule, pos_remaining, neg_remaining)

            if self.verbosity>=3:
                print(f'examples remaining: {len(pos_remaining)} pos, {len(neg_remaining)} neg')
                print()

        # If applicable, trim total conds
        ruleset.trim_conds(max_total_conds=max_total_conds)

        return ruleset

    def _optimize_ruleset(self, ruleset, pos_df, neg_df, prune_size, max_rule_conds=None, random_state=None):
        """ Optimization phase. """

        if self.verbosity>=2:
            print('optimizing ruleset...')
            print()

        pos_remaining = pos_df.copy()
        neg_remaining = neg_df.copy()
        original_ruleset = copy.deepcopy(ruleset)
        if self.verbosity>=4: print('calculate original ruleset potential dl...')
        original_dl = _rs_total_bits(original_ruleset, original_ruleset.possible_conds, pos_df, neg_df, bestsubset_dl=True, verbosity=self.verbosity)
        if self.verbosity>=3:
            print(f'original ruleset potential dl: {rnd(original_dl)}')
            print()
        new_ruleset = copy.deepcopy(ruleset)

        for i, rule in enumerate(original_ruleset.rules):
            pos_growset, pos_pruneset = base.df_shuffled_split(pos_remaining, (1-prune_size), random_state=random_state)
            neg_growset, neg_pruneset = base.df_shuffled_split(neg_remaining, (1-prune_size), random_state=random_state)
            if len(pos_growset) == 0: break # Possible where optimization run > 1

            # Create alternative rules
            if self.verbosity>=4: print(f'creating replacement for {i} of {len(original_ruleset.rules)}: {ruleset.rules[i]}')
            g_replacement = base.grow_rule(pos_growset, neg_growset, original_ruleset.possible_conds, initial_rule=Rule(), max_rule_conds=max_rule_conds, verbosity=self.verbosity)
            replacement_ruleset = Ruleset(base.i_replaced(original_ruleset.rules, i, g_replacement))
            pr_replacement = base.prune_rule(g_replacement, _RIPPER_optimization_prune_metric, pos_pruneset, neg_pruneset, eval_index_on_ruleset=(i,replacement_ruleset), verbosity=self.verbosity)
            replacement_ruleset = Ruleset(base.i_replaced(original_ruleset.rules, i, pr_replacement))
            if self.verbosity>=3:
                print(f'grew replacement {g_replacement}')
                print(f'pruned replacement is {pr_replacement}')

            if self.verbosity>=3: print(f'creating revision for {i} of {len(original_ruleset.rules)}: {ruleset.rules[i]}')
            g_revision = base.grow_rule(pos_growset, neg_growset, original_ruleset.possible_conds, initial_rule=ruleset.rules[i], max_rule_conds=max_rule_conds, verbosity=self.verbosity)
            revision_ruleset = Ruleset(base.i_replaced(original_ruleset.rules, i, g_revision))
            pr_revision = base.prune_rule(g_revision, _RIPPER_optimization_prune_metric, pos_pruneset, neg_pruneset, eval_index_on_ruleset=(i,revision_ruleset), verbosity=self.verbosity)
            revision_ruleset = Ruleset(base.i_replaced(original_ruleset.rules, i, pr_revision))
            if self.verbosity>=3:
                print(f'grew revision {g_replacement}')
                print(f'pruned revision is {pr_replacement}')
                print()

            # Calculate alternative Rulesets' respective lowest potential dls to identify the best version
            if self.verbosity>=3: print(f'calculate potential dl for ds with replacement {pr_replacement}')
            replacement_dl = _rs_total_bits(replacement_ruleset, original_ruleset.possible_conds, pos_df, neg_df, bestsubset_dl=True, verbosity=self.verbosity)\
                             if pr_replacement!=rule else original_dl
            if self.verbosity>=3: print(f'calculate potential dl for ds with revision {pr_revision}')
            revision_dl = _rs_total_bits(revision_ruleset, original_ruleset.possible_conds, pos_df, neg_df, bestsubset_dl=True, verbosity=self.verbosity)\
                          if pr_revision!=rule else original_dl
            best_rule = [rule, pr_replacement, pr_revision][base.argmin([original_dl, replacement_dl, revision_dl])]

            if self.verbosity>=2:
                print(f'\nrule {i+1} of {len(original_ruleset.rules)}')
                rep_str = pr_replacement.__str__() if pr_replacement!=rule else 'unchanged'
                rev_str = pr_revision.__str__() if pr_revision!=rule else 'unchanged'
                best_str = best_rule.__str__() if best_rule!=rule else 'unchanged'
                if self.verbosity==2:
                    print(f'original: {rule}')
                    print(f'replacement: {rep_str}')
                    print(f'revision: {rev_str}')
                    print(f'*best: {best_str}')
                    if best_rule in new_ruleset: print(f'best already included in optimization -- retaining original')
                    print()
                else:
                    print(f'original: {rule}) | {rnd(original_dl)} bits')
                    print(f'replacement: {rep_str} | {rnd(replacement_dl)} bits')
                    print(f'revision: {rev_str} | {rnd(revision_dl)} bits')
                    print(f'*best: {best_str} | {rnd(min([replacement_dl, revision_dl, original_dl]))} bits')
                    if best_rule in new_ruleset: print(f'best already included in optimization -- retaining original')
                    print()
            if best_rule not in new_ruleset:
                new_ruleset.rules[i] = best_rule
            else:
                new_ruleset.rules[i] = rule

            # Remove covered examples
            pos_remaining, neg_remaining = base.rm_covered(rule, pos_remaining, neg_remaining)
            if self.verbosity>=3:
                print(f'examples remaining: {len(pos_remaining)} pos, {len(neg_remaining)} neg')
                print()

            # If there are no pos data remaining to train optimization (could happen if optimization run >1), keep remaining rules the same
            if len(pos_remaining)==0: break

        return new_ruleset

    def _set_possible_conds(self, df):
        self.possible_conds = []
        for feat in df.columns.values:
            for val in df[feat].unique():
                self.possible_conds.append(Cond(feat, val))

#### HELPER Class #####

class RulesetStats:
    # This class is not used in the current implementation but could come in handy for future optimization
    # by storing and retreiving calculations that may be repeated.
    # Haven't incorporated it because there are bigger fish to fry, optimization-wise.
    def __init__(self, ruleset=None):
        self.subset_dls = []
        self.ruleset = Ruleset() if ruleset is None else ruleset
        self.dl = 0

    def update(self, ruleset, possible_conds, pos_df, neg_df, verbosity=0):

        # Find first mismatching rule index
        # If there is no mismatch, return
        # If there is a mismatch, update self.ruleset and wipe subsequent dls
        index = 0
        while index < len(self.ruleset) and \
              index < len(ruleset) and \
              self.ruleset[index] == ruleset[index]:
            index += 1
        if index == len(ruleset) and index == len(self.ruleset):
            if verbosity>=4: print(f'not updating stats -- no ruleset change found')
            return
        if verbosity>=4: print(f'updating stats from index {index}')
        self.ruleset.rules[index:] = ruleset[index:]
        self.subset_dls = self.subset_dls[:index]

        # Beginning with index, update subset dls
        for i in range(index, len(ruleset)):
            rule = ruleset[i]
            subset = Ruleset(ruleset.rules[:i+1])
            subset_dl = _rs_total_bits(subset, possible_conds, pos_df, neg_df, verbosity=verbosity)
            self.subset_dls.append(subset_dl)

        self.dl = self.subset_dls[-1]

    def dl_change(self, index):
        return self.subset_dls[index] - self.subset_dls[index-1]

    def potential_dl_stats(self, possible_conds, pos_df, neg_df,
                          ret_ruleset=True, ret_dl=False, verbosity=0):
        if not any((ret_ruleset, ret_dl)):
            raise ValueError('method dl_pruned_ruleset called without any return values specified')

        tempStats = copy.deepcopy(self)
        i = len(tempStats.ruleset) - 1
        while i > 0:
            if tempStats.dl_change(i) > 0:
                if verbosity>=4: print(f'rule {i} raised dl -- removing')
                tempStats.update(Ruleset(tempStats.ruleset[:i]+tempStats.ruleset[i+1:]), possible_conds, pos_df, neg_df)
                if verbosity>=4: print(f'new ruleset is {tempStats.ruleset}')
            i -= 1

        return tempStats


###################################
##### RIPPER-specific Metrics #####
###################################

def _RIPPER_growphase_prune_metric(rule, pos_pruneset, neg_pruneset):
    """ RIPPER/IREP* prune metric.
        Returns the prune value of a candidate Rule.

        Cohen's formula is (p-n) / (p+n).
        Unclear from the paper how they handle divzero (where p+n=0), so I Laplaced it.
        Weka's solution was to modify the formula to (p+1)/(p+n+2), but the (non-NaN) values I got appeared closer to those of the original formula.
    """
    # I imagine Weka's is 1/2 because that's closer to a 50-50 class distribution?
    p = rule.num_covered(pos_pruneset)
    n = rule.num_covered(neg_pruneset)
    return (p - n + 1) / (p + n + 1)

def _RIPPER_optimization_prune_metric(rule, pos_pruneset, neg_pruneset):
    return base.accuracy(rule, pos_pruneset, neg_pruneset)

def _r_theory_bits(rule, possible_conds, bits_dict=None, verbosity=0):
    """ Returns description length (in bits) for a single Rule. """

    if hasattr(rule, 'dl'):
        return rule.dl
    else:
        if type(rule) != Rule:
            raise TypeError(f'param rule in _r_theory_bits is type {type(rule)}; it should be type Rule')
        k = len(rule.conds)                                 # Number of rule conditions
        n = len(possible_conds)                             # Number of possible conditions
        pr = k/n

        S = k*math.log2(1/pr) + (n-k)*math.log2((1/(1-pr)))  # S(n, k, pr)
        K = math.log2(k)                                     # Number bits need to send integer k
        rule_dl = 0.5*(K + S)                                # Divide by 2 a la Quinlan. Cohen: "to adjust for possible redundency in attributes"
        if verbosity>=5: print(f'rule theory bits| {rule} k {k} n {n} pr {rnd(pr)}: {rnd(rule_dl)} bits')

        #rule.dl = rule_dl
        return rule_dl

def _rs_theory_bits(ruleset, possible_conds, pos_df, neg_df, verbosity=0):
    """ Returns theory description length (in bits) for a Ruleset. """

    if type(ruleset) != Ruleset:
        raise TypeError(f'param ruleset in _rs_theory_bits should be type Ruleset')
    """ Returns sum of theory bits for each Rule in ruleset """
    total = 0
    for rule in ruleset.rules:
        total += _r_theory_bits(rule, possible_conds, verbosity=verbosity)
        #total += rule_bits(rule, possible_conds, rem_pos, rem_neg, verbosity=verbosity)
        #rem_pos, rem_neg = base.rm_covered(rule, rem_pos, rem_neg)
    if verbosity>=5: print(f'ruleset theory bits| {rnd(total)}')

    #ruleset.dl = total
    return total

def _exceptions_bits(ruleset, pos_df, neg_df, verbosity=0):
    """ Returns description length (in bits) for exceptions to a Ruleset's coverage. """

    if type(ruleset) != Ruleset:
        raise TypeError(f'to avoid double-counting, _exceptions_bits should calculate exceptions over entire set of rules with type Ruleset')
    N = len(pos_df) + len(neg_df)                                 # Total number of examples
    p = ruleset.num_covered(pos_df) + ruleset.num_covered(neg_df) # Total number of examples classified as positive = total covered
    fp = ruleset.num_covered(neg_df)                              # Number false positives = negatives covered by the ruleset
    fn = len(pos_df) - ruleset.num_covered(pos_df)                # Number false negatives = positives not covered by the ruleset
    exceptions_dl = math.log2(base.nCr(p,fp)) + math.log2(base.nCr((N-p),fn))
    if verbosity>=5: print(f'exceptions_bits| {ruleset.truncstr()}: \n N {N} p {p} fp {fp} fn {fn}: exceptions_bits {rnd(exceptions_dl)}')

    return exceptions_dl

def _rs_total_bits(ruleset, possible_conds, pos_df, neg_df, bestsubset_dl=False, ret_bestsubset=False, verbosity=0):
    """ Returns total description length (in bits) of ruleset -- the sum of its theory dl and exceptions dl.

        bestsubset_dl (optional, <bool>): whether to return estimated minimum possible dl were all rules that increase dl to be removed
        ret_bestsubset (optional): whether to return the best subset that was found. Return format will be (<Ruleset>,dl).
    """

    # The RIPPER paper is brief and unclear w/r how to evaluate a ruleset for best potential dl.

    # 1) Do you reevaluate already-visited rules or evaluate each rule independently of one another?
    # Weka's source code comments that you are not supposed to, and that this is "bizarre."
    # Perhaps not recursing so -- and getting a possibly sub-optimal mdl -- could be viewed as a greedy time-saver?
    # After all, this is supposed to be an iterative algorithm, it could optimize more times with future k's,
    # and it's not like we're performing an exhaustive search of every possible combination anyways.

    # 2) In what order are you supposed to evaluate? FIFO or LIFO?
    # Footnote 7 suggests optimization is done FIFO; the previous page suggests IREP* final dl reduction is done LIFO;
    # and context suggests dl reduction should be performed the same way both times.

    # I chose to greedy for #1, and FIFO for #2 but may choose differently in a future version if it seems more appropriate.
    # In any case, RIPPER's strong performance on the test sets vs. RandomForest suggests it may not matter all that much.

    if type(ruleset) != Ruleset:
        raise TypeError(f'param ruleset in _rs_total_bits should be type Ruleset')
    if ret_bestsubset and not bestsubset_dl:
        raise ValueError(f'bestsubset_dl must be True in order to return bestsubset_dl')

    if not bestsubset_dl:
        theory_bits = _rs_theory_bits(ruleset, possible_conds, pos_df, neg_df, verbosity=verbosity)
        data_bits = _exceptions_bits(ruleset, pos_df, neg_df, verbosity=verbosity)
        if verbosity>=3: print(f'total ruleset bits | {rnd(theory_bits + data_bits)}')
        return theory_bits + data_bits
    else:
        # Collect the dl of each subset
        subset_dls = []
        theory_dl = 0
        if verbosity>=5: print(f'find best potential dl for {ruleset}:')
        for i, rule in enumerate(ruleset.rules): # Separating theory and exceptions dls in this way means you don't have to recalculate theory each time
            subset = Ruleset(ruleset.rules[:i+1])
            rule_theory_dl = _r_theory_bits(rule, possible_conds, verbosity=verbosity)
            theory_dl += rule_theory_dl
            exceptions_dl = _exceptions_bits(subset, pos_df, neg_df, verbosity=verbosity)
            subset_dls.append(theory_dl + exceptions_dl)
            if verbosity>=5: print(f'subset 0-{i} | dl: {rnd(subset_dls[i])}')

        # Build up the best Ruleset and calculate the mdl
        mdl_ruleset = Ruleset()
        for i, rule, in enumerate(ruleset.rules):
            if i==0 or subset_dls[i] <= subset_dls[i-1]: # Rule i does not worsen the dl
                mdl_ruleset.add(rule)
        if verbosity>=5:
            print(f'subset dls: {[(i,rnd(dl)) for i,dl in enumerate(subset_dls)]}')
            print(f'best potential ruleset: {mdl_ruleset}')
        mdl = _rs_total_bits(mdl_ruleset, possible_conds, pos_df, neg_df, bestsubset_dl=False, verbosity=0) # About to print value below
        if verbosity>=5:
            print(f'best potential dl was {rnd(mdl)}')
            print()
        if not ret_bestsubset:
            return mdl
        else:
            return (mdl_ruleset, mdl)



import pytest
import pandas as pd

from wittgenstein.base import pos, neg, pos_neg_split, df_shuffled_split, num_pos, num_neg
from wittgenstein.base import Cond, Rule, Ruleset

FIRST_10_EXAMPLES = pd.read_csv('tests/house-votes-84.csv').head(10)
CLASS_FEAT = 'Party'
POS_CLASS = 'democrat'
SPLIT_SIZE = .6

FEATURE1, VALUE1 = 'Handicapped-infants', 'n'
FEATURE2, VALUE2 = 'Water-project-cost-sharing', 'y'
FEATURE3, VALUE3 = 'anti-satellite-test-ban', 'n'
FEATURE4, VALUE4 = 'el-salvador-aid', 'y'
COND1 = Cond(FEATURE1, VALUE1)
COND2 = Cond(FEATURE2, VALUE2)
COND3 = Cond(FEATURE3, VALUE3)
COND4 = Cond(FEATURE4, VALUE4)

##### Test Helpers ######

def test_len_pos_df_is_6():
    assert len(pos(FIRST_10_EXAMPLES,CLASS_FEAT,POS_CLASS))==6

def test_len_neg_df_is_4():
    assert len(neg(FIRST_10_EXAMPLES,CLASS_FEAT,POS_CLASS))==4

def test_split_pos_is_6_neg_is_4():
    pos_df, neg_df = pos_neg_split(FIRST_10_EXAMPLES,CLASS_FEAT,POS_CLASS)
    assert(len(pos_df),len(neg_df))==(6,4)

def test_shuffled_splits_are_len_7_len_3():
    df1, df2 = df_shuffled_split(FIRST_10_EXAMPLES,.7,random_state=None)
    assert(len(df1),len(df2))==(7,3)

def test_num_pos_is_6_num_neg_is_4():
    p = num_pos(FIRST_10_EXAMPLES, CLASS_FEAT, POS_CLASS)
    n = num_neg(FIRST_10_EXAMPLES, CLASS_FEAT, POS_CLASS)
    assert((p,n)==(6,4))

def test_num_pos_is_0_num_neg_is_0():
    pos_df, neg_df = pos_neg_split(FIRST_10_EXAMPLES,CLASS_FEAT,POS_CLASS)
    p = num_pos(neg_df, CLASS_FEAT, POS_CLASS)
    n = num_neg(pos_df, CLASS_FEAT, POS_CLASS)
    assert((p,n)==(0,0))

####### Test Conds ######

def test_conds_are_equal():
    assert(Cond('feature','value')==Cond('feature','value'))

def test_conds_are_unequal_value():
    assert(Cond('feature','value1')!=Cond('feature','value2'))

def test_conds_are_unequal_feature():
    assert(Cond('feature1','value')!=Cond('feature2','value'))

def test_cond_covers():
    cond = Cond(FEATURE1, VALUE1)
    assert(set(cond.covers(FIRST_10_EXAMPLES)[FEATURE1].tolist()) == {VALUE1})

def test_cond_doesnt_cover():
    cond = Cond(FEATURE1, VALUE1)
    assert(len(cond.covers(FIRST_10_EXAMPLES[FIRST_10_EXAMPLES[FEATURE1] != VALUE1])) == 0)

def test_cond_num_covered_is_7():
    cond = Cond(FEATURE1, VALUE1)
    assert(len(cond.covers(FIRST_10_EXAMPLES)[FEATURE1].tolist()) == 7)

##### Test Rule #####

def test_empty_rules_equal():
    assert(Rule() == Rule())

def test_rules_len_1_equal():
    assert(Rule([COND1]) == Rule([COND1]))

def test_rules_len_1_unequal():
    assert(Rule([COND1]) != Rule([COND2]))

def test_rules_len_2_equal():
    rule1 = Rule([COND1, COND2])
    rule2 = Rule([COND1, COND2])
    assert(rule1 == rule2)

def test_disordered_rules_len_2_equal():
    rule1 = Rule([COND1, COND2])
    rule2 = Rule([COND2, COND1])
    assert(rule1 == rule2)

def test_rules_len_2_unequal():
    rule1 = Rule([COND1, COND2])
    rule2 = Rule([COND2, COND3])
    assert(rule1 != rule2)

def test_rule_covers():
    rule = Rule([COND1, COND4])
    rule_covers = rule.covers(FIRST_10_EXAMPLES)
    assertion_df = FIRST_10_EXAMPLES[(FIRST_10_EXAMPLES[FEATURE1]==VALUE1) & (FIRST_10_EXAMPLES[FEATURE4]==VALUE4)]
    assert(set(rule_covers.index.tolist())==set(assertion_df.index.tolist()))

def test_rule_num_covered():
    rule = Rule([COND1, COND4])
    rule_num_covered = rule.num_covered(FIRST_10_EXAMPLES)
    len_assertion_df = len(FIRST_10_EXAMPLES[(FIRST_10_EXAMPLES[FEATURE1]==VALUE1) & (FIRST_10_EXAMPLES[FEATURE4]==VALUE4)])
    assert(rule_num_covered == len_assertion_df)



""" Base classes and functions for ruleset classifiers """

import math
import numpy as np
import operator as op
from functools import reduce
import copy
import warnings
from numpy import var, mean
import time


class Ruleset:
    """ Base Ruleset model.
        Implements collection of Rules in disjunctive normal form.
    """

    def __init__(self, rules=None):
        if rules is None:
            self.rules = []
        else:
            self.rules = rules
        self.cond_count = 0

    def __str__(self):
        return ' V '.join([str(rule) for rule in self.rules])

    def __repr__(self):
        ruleset_str = self.__str__()
        return f'<Ruleset object: {ruleset_str}>'

    def __getitem__(self, index):
        return self.rules[index]

    def __len__(self):
        return len(self.rules)

    def truncstr(self, limit=2, direction='left'):
        """ Return Ruleset string representation limited to a specified number of rules.

            limit: how many rules to return
            direction: which part to return. (valid options: 'left', 'right')
        """
        if len(self.rules)>=limit:
            if direction=='left':
                return Ruleset(self.rules[:limit]).__str__()+'...'
            elif direction=='right':
                return '...'+Ruleset(self.rules[-limit:]).__str__()
            else:
                raise ValueError('direction param must be "left" or "right"')
        else:
            return self.__str__()

    def __eq__(self, other):
        if type(other)!=Ruleset:
            raise TypeError(f'{self} __eq__ {other}: a Ruleset can only be compared with another Ruleset')
        for r in self.rules: # TODO: Ideally, should implement a hash function--in practice speedup would be insignificant
            if r not in other.rules: return False
        for r in other.rules: # Check the other way around too. (Can't compare lengths instead b/c there might be duplicate rules.)
            if r not in self.rules: return False
        return True

    def out_pretty(self):
        """ Prints Ruleset line-by-line. """
        ruleset_str = str([str(rule) for rule in self.rules]).replace(' ','').replace(',',' V\n').replace("'","")
        print(ruleset_str)

    def copy(self, n_rules_limit=None):
        """ Returns a deep copy of self.

            n_rules_limit (optional): keep only this many rules from the original.
        """
        result = copy.deepcopy(self)
        if n_rules_limit is not None:
            result.rules = result.rules[:n_rules_limit]
        return result

    def covers(self, df):
        """ Returns instances covered by the Ruleset. """

        if not self.rules:
            return df
        else:
            covered = self.rules[0].covers(df).copy()
            for rule in self.rules[1:]:
                covered = covered.append(rule.covers(df))
            covered = covered.drop_duplicates()
            return covered

    def num_covered(self, df):
        return len(self.covers(df))

    def add(self, rule):
        self.rules.append(rule)

    def count_rules(self):
        """ Returns number of rules in the Ruleset.
            (For ease of use for users who don't make a habit of directly accessing the Ruleset object.)
        """
        return len(self.rules)

    def count_conds(self):
        """ Returns the total number of conditions in the Ruleset. This is a measurement of complexity that's the conceptual equivalent of counting the nodes in a decision tree. """
        return sum([len(r.conds) for r in self.rules])

    def _set_possible_conds(self, pos_df, neg_df):
        """ Stores a list of all possible conds. """

        #Used in Rule::successors so as not to rebuild it each time,
        # and in exceptions_dl calculations because nCr portion of formula already accounts for no replacement.)

        self.possible_conds = []
        for feat in pos_df.columns.values:
            for val in set(pos_df[feat].unique()).intersection(set(neg_df[feat].unique())):
                self.possible_conds.append(Cond(feat, val))

    def trim_conds(self, max_total_conds=None):
        """ Reduce the total number of Conds in a Ruleset by removing Rules """
        if max_total_conds is not None:
            while len(self.rules) > 0 and self.count_conds() > max_total_conds:
                self.rules.pop(-1)

    def predict(self, X_df, give_reasons=False):
        """ Predict classes of data using a fit Ruleset model.

            args:
                X_df <DataFrame>: examples to make predictions on.

                give_reasons (optional) <bool>: whether to provide reasons for each prediction made.

            returns:
                list of <bool> values corresponding to examples. True indicates positive predicted class; False non-positive class.

                If give_reasons is True, returns a tuple that contains the above list of predictions
                    and a list of the corresponding reasons for each prediction;
                    for each positive prediction, gives a list of all the covering Rules, for negative predictions, an empty list.
        """

        covered_indices = set(self.covers(X_df).index.tolist())
        predictions = [i in covered_indices for i in X_df.index]

        if not give_reasons:
            return predictions
        else:
            reasons = []
            # For each Ruleset-covered example, collect list of every Rule that covers it;
            # for non-covered examples, collect an empty list
            for i, p in zip(X_df.index, predictions):
                example = X_df[X_df.index==i]
                example_reasons = [rule for rule in self.rules if len(rule.covers(example))==1] if p else []
                reasons.append(example_reasons)
            return (predictions, reasons)

    def predict_proba(self, X_df, give_reasons=False, ret_n=False, min_samples=1, discretize=True, bin_transformer=None):
        """ Predict probabilities for each class using a fit Ruleset model.

                args:
                    X_df <DataFrame>: examples to make predictions on.

                    give_reasons (optional) <bool>: whether to provide reasons for each prediction made.
                    min_samples (optional) <int>: return None for each example proba that lack this many samples
                                                  set to None to ignore. (default=None)

                    give_reasons (optional) <bool>: whether to also return reasons for each prediction made.
                    ret_n (optional) <bool>: whether to also return the number of samples used for calculating each examples proba

                returns:
                    numpy array of values corresponding to each example's classes probabilities, or, if give_reasons or ret_n, a tuple containing proba array and list(s) of desired returns

        """

        # Apply binning if necessary
        if discretize and bin_transformer is not None:
            df = X_df.copy()
            df = bin_transform(df, bin_transformer)
        else:
            df = X_df

        # probas for all negative predictions
        uncovered_proba = weighted_avg_freqs([self.uncovered_class_freqs])
        uncovered_n = sum(self.uncovered_class_freqs)

        # make predictions
        predictions, covering_rules = self.predict(df, give_reasons=True)
        N = []

        # collect probas
        probas = np.empty(shape=(len(predictions),uncovered_proba.shape[0]))
        for i, (p, cr) in enumerate(zip(predictions, covering_rules)):
            n = sum([sum(rule.class_freqs) for rule in cr]) # if user requests, check to ensure valid sample size
            if p and (n < 1 or (min_samples and n < min_samples)):
                probas[i, :] = None
                N.append(n)
            elif (not p) and (uncovered_n < 1 or uncovered_n < min_samples):
                probas[i, :] = None
                N.append(n)
            elif p: # pos prediction
                probas[i, :] = weighted_avg_freqs([rule.class_freqs for rule in cr])
                N.append(n)
            else: # neg prediction
                probas[i, :] = uncovered_proba
                N.append(uncovered_n)

        # return probas (and optional extras)
        result = flagged_return([True, give_reasons, ret_n], [probas, covering_rules, N])
        return result

    def recalibrate_proba(self, Xy_df, class_feat, pos_class, min_samples=20, require_min_samples=True, discretize=True, bin_transformer=None):
        """ Recalibrate a Ruleset's probability estimations using unseen labeled data. May improve .predict_proba generalizability.
            Does not affect the underlying model or which predictions it makes -- only probability estimates. Use params min_samples and require_min_samples to select desired behavior.

            Note1: RunTimeWarning will occur as a reminder when min_samples and require_min_samples params might result in unintended effects.
            Note2: It is possible recalibrating could result in some positive .predict predictions with <0.5 .predict_proba positive probability.

            Xy_df <DataFrame>: labeled data

            min_samples <int> (optional): required minimum number of samples per Rule
                                          default=10. set None to ignore min sampling requirement so long as at least one sample exists.
            require_min_samples <bool> (optional): True: halt (with warning) in case min_samples not achieved for all Rules
                                                   False: warn, but still replace Rules that have enough samples
            discretize <bool> (optional): if the classifier has already fit a discretization, automatically discretize recalibrate_proba's training data
                                          default=True
        """

        # If not using min_samples, set it to 1
        if not min_samples or min_samples < 1:
            min_samples = 1

        # Apply binning if necessary
        if discretize and bin_transformer is not None:
            df = Xy_df.copy()
            df = bin_transform(df, bin_transformer)
        else:
            df = Xy_df

        # Collect each Rule's pos and neg frequencies
        rule_class_freqs = [None]*len(self.rules) # to contain new frequencies for each Rule
        insufficient_rules = []
        for i, rule in enumerate(self.rules):
            npos_pred = num_pos(rule.covers(df), class_feat=class_feat, pos_class=pos_class)
            nneg_pred = num_neg(rule.covers(df), class_feat=class_feat, pos_class=pos_class)
            pos_neg_pred = (npos_pred, nneg_pred)
            rule_class_freqs[i] = pos_neg_pred
            if sum(pos_neg_pred) < min_samples:
                insufficient_rules.append(rule)

        # Collect Ruleset's uncovered frequencies
        uncovered = df.drop(self.covers(df).index)
        neg_freq = num_neg(uncovered, class_feat=class_feat, pos_class=pos_class)
        fn_tn = (len(uncovered) - neg_freq, neg_freq)

        if min_samples: # This will always execute b/c None min_samples is set to 1, but it's worth including the logic in case that implementation detail ever changes
            if insufficient_rules:
                warning_str = f'recalibrate_proba: param min_samples={min_samples}; insufficient number of samples from rules {insufficient_rules}'
                warnings.warn(warning_str, RuntimeWarning)
            if sum(fn_tn) < min_samples:
                warning_str = f'recalibrate_proba: param min_samples={min_samples}; insufficient number of negatively labled samples {sum(fn_tn)}'
                warnings.warn(warning_str, RuntimeWarning)
            if insufficient_rules or sum(fn_tn) < min_samples:
                if require_min_samples:
                    warning_str = f'recalibrate_proba: recalibrating halted. to recalibrate, try lowering min_samples or set require_min_samples to False'
                    warnings.warn(warning_str, RuntimeWarning)
                    return
                else:
                    warning_str = f'recalibrate_proba: retaining probabilities for rules without enough samples and because param required_min_samples=False; recalibrate any rules that have samples > parm min_samples={min_samples}'
                    warnings.warn(warning_str, RuntimeWarning)

        # Assign collected frequencies to Rules
        for rule, freqs in zip(self.rules, rule_class_freqs):
            if freqs is not None:
                rule.class_freqs = freqs
            else:
                pass # ignore Rules that don't have enough samples

        # Assign Ruleset's uncovered frequencies
        self.uncovered_class_freqs = fn_tn

class Rule:
    """ Class implementing conjunctions of Conds """

    def __init__(self, conds=None):
        if conds is None:
            self.conds = []
        else:
            self.conds = conds

    def __str__(self):
        rule_str = str([str(cond) for cond in self.conds]).replace(',','^').replace("'","").replace(' ','')
        return rule_str

    def __repr__(self):
        rule_str = str([str(cond) for cond in self.conds]).replace(', ','^').replace("'","").replace(' ','')
        return f'<Rule object: {rule_str}>'

    def __add__(self, cond):
        if type(cond)==Cond:
            return Rule(self.conds+[cond])
        else:
            raise TypeError(f'{self} + {cond}: Rule objects can only conjoin Cond objects.')

    def __eq__(self, other):
        if type(other)!=Rule:
            raise TypeError(f'{self} __eq__ {other}: a Rule can only be compared with another rule')
        if len(self.conds)!=len(other.conds): return False
        return set([str(cond) for cond in self.conds]) == set([str(cond) for cond in other.conds])

    def isempty(self):
        return len(self.conds)==0

    def covers(self, df):
        """ Returns instances covered by the Rule. """
        covered = df.head(len(df))
        for cond in self.conds:
            covered = cond.covers(covered)
        return covered

    def num_covered(self, df):
        return len(self.covers(df))

    def covered_feats(self):
        """ Returns list of features covered by the Rule """
        return [cond.feature for cond in self.conds]

    #############################################
    ##### Rule::grow/prune helper functions #####
    #############################################

    def successors(self, possible_conds, pos_df, neg_df):
        """ Returns a list of all valid successor rules.

        possible_conds: list of Conds to consider conjoining to create successors.
                        passing None defaults to create this param from pos_df and neg_df --
                        however, if pos_df and neg_df are data subsets, it will only generate possible_conds
                        from their available values.
        """

        if possible_conds is not None:
            successor_conds = [cond for cond in possible_conds if cond not in self.conds]
            return [Rule(self.conds+[cond]) for cond in successor_conds]
        else:
            successor_rules = []
            for feat in pos_df.columns.values:
                for val in set(pos_df[feat].unique()).intersection(set(neg_df[feat].unique())):
                    if feat not in self.covered_feats(): # Conds already in Rule and Conds that contradict Rule aren't valid successors / NB Rules are short; this is unlikely to be worth the overhead of cheacking
                        successor_rules.append(self+Cond(feat, val))
            return successor_rules

class Cond:
    """ Class implementing conditional. """

    def __init__(self, feature, val):
        self.feature = feature
        self.val = val

    def __str__(self):
        return f'{self.feature}={self.val}'

    def __repr__(self):
        return f'<Cond object: {self.feature}={self.val}>'

    def __eq__(self, other):
        return self.feature == other.feature and self.val == other.val

    def __hash__(self):
        return hash((self.feature, self.val))

    def covers(self, df):
        """ Returns instances covered by the Cond (i.e. those which are not in contradiction with it). """
        return df[df[self.feature]==self.val]

    def num_covered(self, df):
        return len(self.covers(df))

##########################
##### BASE FUNCTIONS #####
##########################

def grow_rule(pos_df, neg_df, possible_conds, initial_rule=Rule(), max_rule_conds=None, verbosity=0):
    """ Fit a new rule to add to a ruleset """
    # Possible optimization: remove data after each added cond?

    #print('grow_rule original')

    rule0 = copy.deepcopy(initial_rule)
    if verbosity>=4:
        print(f'growing rule from initial rule: {rule0}')

    rule1 = copy.deepcopy(rule0)
    while (len(rule0.covers(neg_df)) > 0 and rule1 is not None) and (max_rule_conds is None or len(rule1.conds) < max_rule_conds): # Stop refining rule if no negative examples remain
        rule1 = best_successor(rule0, possible_conds, pos_df, neg_df, verbosity=verbosity)
        #print(f'growing rule... {rule1}')
        if rule1 is not None:
            rule0 = rule1
            if verbosity>=4:
                print(f'negs remaining {len(rule0.covers(neg_df))}')

    if not rule0.isempty():
        if verbosity>=2: print(f'grew rule: {rule0}')
        return rule0
    else:
        #warnings.warn(f"grew an empty rule {rule0} over {len(pos_df)} pos and {len(neg_df)} neg", RuntimeWarning)#, stacklevel=1, source=None)
        return rule0

def prune_rule(rule, prune_metric, pos_pruneset, neg_pruneset, eval_index_on_ruleset=None, verbosity=0):
    """ Returns a pruned version of the Rule by removing Conds

        rule: Rule to prune
        prune_metric: function that returns value to maximize
        pos_pruneset: df of positive class examples
        neg_pruneset: df of non-positive class examples

        eval_index_on_ruleset (optional): tuple(rule_index, ruleset)
            pass the rest of the Rule's Ruleset (excluding the Rule in question),
            in order to prune the rule based on the performance of its entire Ruleset,
            rather than on the rule alone. For use during optimization stage.
    """

    if rule.isempty():
        warnings.warn(f"can't prune empty rule {rule}", RuntimeWarning)#, stacklevel=1, source=None)
        return rule

    if not eval_index_on_ruleset:

        # Currently-best pruned rule and its prune value
        best_rule = copy.deepcopy(rule)
        best_v = 0

        # Iterative test rule
        current_rule = copy.deepcopy(rule)

        while current_rule.conds:
            v = prune_metric(current_rule, pos_pruneset, neg_pruneset)
            if verbosity>=5: print(f'prune value of {current_rule}: {rnd(v)}')
            if v is None:
                return None
            if v >= best_v:
                best_v = v
                best_rule = copy.deepcopy(current_rule)
            current_rule.conds.pop(-1)

        if verbosity>=2:
            if len(best_rule.conds)!=len(rule.conds):
                print(f'pruned rule: {best_rule}')
            else:
                print(f'pruned rule unchanged')
        return best_rule

    else:
        # Check if index matches rule to prune
        rule_index, ruleset = eval_index_on_ruleset
        if ruleset.rules[rule_index] != rule:
            raise ValueError(f'rule mismatch: {rule} - {ruleset.rules[rule_index]} in {ruleset}')

        current_ruleset = copy.deepcopy(ruleset)
        current_rule = current_ruleset.rules[rule_index]
        best_ruleset = copy.deepcopy(current_ruleset)
        best_v = 0

        # Iteratively prune and test rule over ruleset.
        # This is unfortunately expensive.
        while current_rule.conds:
            v = prune_metric(current_ruleset, pos_pruneset, neg_pruneset)
            if verbosity>=5: print(f'prune value of {current_rule}: {rnd(v)}')
            if v is None:
                return None
            if v >= best_v:
                best_v = v
                best_rule = copy.deepcopy(current_rule)
                best_ruleset = copy.deepcopy(current_ruleset)
            current_rule.conds.pop(-1)
            current_ruleset.rules[rule_index] = current_rule
        return best_rule

class Timer:
    """ Simple, useful class for keeping track of how long something has been taking """

    def __init__(self):
        """ Create Timer object and hit start. """

        self.start = time.time()

    def buzz(self, reset=True):
        """ Returns time elapsed since Timer was created or reset, in seconds.

            args:
                reset (optional): whether to reset the clock.
        """

        last_buzz = self.start
        now = time.time()
        if reset:
            self.start = now
        return(str(int(now-last_buzz)))

    def stop(self):
        """ Freeze the clock at the amount of time elapsed since Timer was created or reset. """

        self.elapsed = time.time()-self.start

    ###################
    ##### METRICS #####
    ###################

def gain(before, after, pos_df, neg_df):
    """ Returns the information gain from self to other """

    p0count = before.num_covered(pos_df)
    p1count = after.num_covered(pos_df)
    n0count = before.num_covered(neg_df)
    n1count = after.num_covered(neg_df)
    return p1count * (math.log2((p1count + 1) / (p1count + n1count + 1)) - math.log2((p0count + 1) / (p0count + n0count + 1)))

def precision(object, pos_df, neg_df):
    """ Returns precision value of object's classification.
        object: Cond, Rule, or Ruleset
    """

    pos_covered = object.covers(pos_df)
    neg_covered = object.covers(neg_df)
    total_n_covered = len(pos_covered)+len(neg_covered)
    if total_n_covered == 0:
        return None
    else:
        return len(pos_covered) / total_n_covered

def score_accuracy(predictions, actuals):
    """ For evaluating trained model on test set.

        predictions: <iterable<bool>> True for predicted positive class, False otherwise
        actuals:     <iterable<bool>> True for actual positive class, False otherwise
    """
    t = [pr for pr,act in zip(predictions,actuals) if pr==act]
    n = predictions
    return len(t)/len(n)

def accuracy(object, pos_pruneset, neg_pruneset):
    """ Returns accuracy value of object's classification.
        object: Cond, Rule, or Ruleset
    """
    P = len(pos_pruneset)
    N = len(neg_pruneset)
    if P + N == 0:
        return None

    tp = len(object.covers(pos_pruneset))
    tn = N - len(object.covers(neg_pruneset))
    return (tp + tn) / (P + N)

def best_successor(rule, possible_conds, pos_df, neg_df, verbosity=0):
    """ Returns for a Rule its best successor Rule according to FOIL information gain metric.

        eval_on_ruleset: option to evaluate gain with extra disjoined rules (for use with RIPPER's post-optimization)
    """

    best_gain = 0
    best_successor_rule = None

    for successor in rule.successors(possible_conds, pos_df, neg_df):
        g = gain(rule, successor, pos_df, neg_df)
        if g > best_gain:
            best_gain = g
            best_successor_rule = successor
    if verbosity>=5: print(f'gain {rnd(best_gain)} {best_successor_rule}')
    return best_successor_rule

    ###################
    ##### HELPERS #####
    ###################

def pos_neg_split(df, class_feat, pos_class):
    """ Split df into pos and neg classes. """
    pos_df = pos(df, class_feat, pos_class)
    neg_df = neg(df, class_feat, pos_class)
    return pos_df, neg_df

def df_shuffled_split(df, split_size, random_state=None):
    """ Returns tuple of shuffled and split DataFrame.
        split_size: proportion of rows to include in tuple[0]
    """
    df = df.sample(frac=1, random_state=random_state).reset_index(drop=True)
    split_at = int(len(df)*split_size)
    return (df[:split_at], df[split_at:])

def pos(df, class_feat, pos_class):
    """ Returns subset of instances that are labeled positive. """
    #""" Returns X,y subset that are labeled positive """
    return df[df[class_feat] == pos_class]
    #return [(Xi, yi) for Xi, yi in zip(X, y) if y==pos_class]

def neg(df, class_feat, pos_class):
    """ Returns subset of instances that are NOT labeled positive. """
    #""" Returns X,y subset that are NOT labeled positive """
    return df[df[class_feat] != pos_class]
    #return [(Xi, yi) for Xi, yi in zip(X, y) if y!=pos_class]

def num_pos(df, class_feat, pos_class):
    """ Returns number of instances that are labeled positive. """
    #""" Returns X,y subset that are labeled positive """
    return len(df[df[class_feat] == pos_class])
    #return len(_pos(X, y, pos_class))

def num_neg(df, class_feat, pos_class):
    """ Returns number of instances that are NOT labeled positive. """
    #""" Returns X,y subset that are NOT labeled positive """
    return len(df[df[class_feat] != pos_class])
    #return len(_neg(X, y, pos_class))

def nCr(n, r):
    """ Returns number of combinations C(n, r) """
    def product(numbers):
        return reduce(op.mul, numbers, 1)

    num = product(range(n, n-r, -1))
    den = product(range(1, r+1))
    return num//den

def rnd(float, places='default'):
    """ places: number of decimal places to round.
                set to 'default': defaults to 1 decimal place if float < 100, otherwise defaults to 0 places
    """
    if places=='default':
        if float<1:
            places = 2
        elif float<100:
            places = 1
        else:
            places = 0
    rounded = round(float, places)
    if rounded!=int(rounded):
        return rounded
    else:
        return int(rounded)

def argmin(list_):
    """ Returns index of minimum value. """
    lowest_val = list_[0]
    lowest_i = 0
    for i, val in enumerate(list_):
        if val < lowest_val:
            lowest_val = val
            lowest_i = i
    return lowest_i

def i_replaced(list_, i, value):
    """ Returns a new list with element i replaced by value.
        Pass None to value to return list with element i removed.
    """
    if value is not None:
        return list_[:i]+[value]+list_[i+1:]
    else:
        return list_[:i]+list_[i+1:]

def rm_covered(object, pos_df, neg_df):
    """ Return pos and neg dfs of examples that are not covered by object """
    return (pos_df.drop(object.covers(pos_df).index, axis=0, inplace=False),\
            neg_df.drop(object.covers(neg_df).index, axis=0, inplace=False))

def trainset_classfeat_posclass(df, y=None, class_feat=None, pos_class=None):
    """ Process params into trainset, class feature name, and pos class, for use in .fit methods. """

    # Ensure class feature is provided
    if y is None and class_feat is None:
        raise ValueError('y or class_feat argument is required')

    # Ensure no class feature name mismatch
    if y is not None and class_feat is not None \
            and hasattr(y, 'name') \
            and y.name != class_feat:
        raise ValueError(f'Value mismatch between params y {y.name} and class_feat {class_feat}. Besides, you only need to provide one of them.')

    # Set class feature name
    if class_feat is not None:
        # (IOW, pass)
        class_feat = class_feat
    elif y is not None and hasattr(y, 'name'):
        # If y is a pandas Series, try to get its name
        class_feat = y.name
    else:
        # Create a name for it
        class_feat = 'Class'

    # If necessary, merge y into df
    if y is not None:
        df[class_feat] = y

    # If provided, define positive class name. Otherwise, assign one.
    if pos_class is not None:
        pos_class = pos_class
    else:
        pos_class = df.iloc[0][class_feat]

    return (df, class_feat, pos_class)

def weighted_avg_freqs(counts):
    """ Return weighted mean proportions of counts in the list

        counts <list<tuple>>
    """
    arr = np.array(counts)
    total = arr.flatten().sum()
    return arr.sum(axis=0) / total

def flagged_return(flags, objects):
    """ Returns only objects with corresponding True flags
        Useful when """
    if sum(flags)==1:
        return objects[0]
    elif sum(flags)>1:
        return tuple([object for flag, object in zip(flags, objects) if flag])
    else:
        return ()

def as_type(df, new_type, ignore_cols=[]):
    result = df.copy()
    ignore = set(ignore_cols)
    for col in df.columns:
        if col not in ignore:
            result[col] = result[col].astype(new_type)
    return result

########################################
##### BONUS: FUNCTIONS FOR BINNING #####
########################################

def bin_df(df, n_discretize_bins=10, ignore_feats=[], verbosity=0):
    """ Returns df with seemingly numeric features binned, and the bin_transformer, if binning takes places. """

    isbinned = False
    min_unique = n_discretize_bins if n_discretize_bins else 10 # Minimum unique values to consider continuous. Override with 10 (rather than halt) to allow warnings to occur
    numeric_feats = find_numeric_feats(df, min_unique=min_unique, ignore_feats=ignore_feats)

    if numeric_feats:
        if n_discretize_bins:
            if verbosity==1:
                print(f'binning data...\n')
            elif verbosity>=2:
                print(f'binning features {numeric_feats}...')
            binned_df = df.copy()
            bin_transformer = fit_bins(binned_df, n_bins=n_discretize_bins, output=False, ignore_feats=ignore_feats, verbosity=verbosity)
            binned_df = bin_transform(binned_df, bin_transformer)
            isbinned = True
        else:
            n_unique_values = sum([len(u) for u in [df[f].unique() for f in numeric_feats]])
            warnings.warn(f"There are {len(numeric_feats)} apparent numeric features: {numeric_feats}. \n Treating {n_unique_values} numeric values as nominal. To auto-discretize features, assign a value to parameter 'n_discretize_bins.'", RuntimeWarning)

    if isbinned:
        return binned_df, bin_transformer
    else:
        return df, None

def find_numeric_feats(df, min_unique=10, ignore_feats=[]):
    """ Returns df features that seem to be numeric. """
    #feats = df.dtypes[(df.dtypes=='float64') | (df.dtypes=='int64')].index.tolist()
    feats = df._get_numeric_data().columns
    feats = [f for f in feats if f not in ignore_feats]
    feats = [f for f in feats if len(df[f].unique())>min_unique]
    return feats

def fit_bins(df, n_bins=5, output=False, ignore_feats=[], verbosity=0):
    """
    Returns a dict definings fits for numerical features
    A fit is an ordered list of tuples defining each bin's range

    Returned dict allows for fitting to training data and applying the same fit to test data
    to avoid information leak.
    """

    def bin_fit_feat(df, feat, n_bins=10):
        """
        Returns list of tuples defining bin ranges for a numerical feature
        """
        n_bins = min(n_bins, len(df[feat].unique())) # In case there are fewer unique values than n_bins
        bin_size = len(df)//n_bins
        sorted_df = df.sort_values(by=[feat])
        sorted_values = sorted_df[feat].tolist()

        if verbosity>=4: print (f'{feat}: fitting {len(df[feat].unique())} unique vals into {n_bins} bins')
        bin_ranges = []
        finish_i=-1
        sizes=[]

        bin = 0
        start_i = 0
        while bin < n_bins and start_i < len(sorted_values):
            finish_i = min(start_i+bin_size, len(sorted_df)-1)
            while finish_i<len(sorted_df)-1 and finish_i!=0 and \
                    sorted_df.iloc[finish_i][feat]==sorted_df.iloc[finish_i-1][feat]: # ensure next bin begins on a new value
                finish_i+=1
                #print(f'{sorted_df.iloc[finish_i][feat]} {sorted_df.iloc[finish_i-1][feat]} {finish_i}')
            sizes.append(finish_i-start_i)
            if verbosity >=4: print(f'bin #{bin}, start_idx {start_i} value: {sorted_df.iloc[start_i][feat]}, finish_idxx {finish_i} value: {sorted_df.iloc[finish_i][feat]}')
            start_val = sorted_values[start_i]
            finish_val = sorted_values[finish_i]
            bin_range = (start_val, finish_val)
            bin_ranges.append(bin_range)

            start_i = finish_i+1
            bin += 1

        if verbosity>=4: print(f'-bin sizes {sizes}; dataVMR={rnd(var(df[feat])/mean(df[feat]))}, binVMR={rnd(var(sizes)/mean(sizes))}')#, axis=None, dtype=None, out=None, ddof=0)})
        return bin_ranges

    # Create dict to store fit definitions for each feature
    fit_dict = {}
    feats_to_fit = find_numeric_feats(df,ignore_feats=ignore_feats)
    if verbosity==2: print(f'fitting bins for features {feats_to_fit}')
    if verbosity>=2: print()

    # Collect fits in dict
    count=1
    for feat in feats_to_fit:
        fit = bin_fit_feat(df, feat, n_bins=n_bins)
        fit_dict[feat] = fit
    return fit_dict

def bin_transform(df, fit_dict, names_precision=2):
    """
    Uses a pre-collected dictionary of fits to transform df features into bins.
    Returns the fit df rather than modifying inplace.
    """

    def bin_transform_feat(df, feat, bin_fits, names_precision=names_precision):
        """
        Returns new dataframe with n_bin bins replacing each numerical feature
        """

        def renamed(bin_fit_list, value, names_precision=names_precision):
            """
            Returns bin string name for a given numberical value
            Assumes bin_fit_list is ordered
            """
            min_val = bin_fit_list[0][0]
            max_val = bin_fit_list[-1][1]
            for bin_fit in bin_fits:
                if value <= bin_fit[1]:
                    start_name = str(round(bin_fit[0], names_precision))
                    finish_name = str(round(bin_fit[1], names_precision))
                    bin_name = '-'.join([start_name, finish_name])
                    return bin_name
            if value < min_val:
                return min_val
            elif value > max_val:
                return max_val
            else:
                raise ValueError('No bin found for value', value)

        renamed_values = []
        for value in df[feat]:
            bin_name = renamed(bin_fits, value, names_precision)
            renamed_values.append(bin_name)

        return renamed_values

    # Replace each feature with bin transformations
    for feat, bin_fits in fit_dict.items():
        feat_transformation = bin_transform_feat(df, feat, bin_fits, names_precision=names_precision)
        df[feat] = feat_transformation
    return df



import setuptools
setuptools.setup(
  name = 'wittgenstein',
  version = '0.1.6',
  license='MIT',
  description = 'Ruleset covering algorithms for explainable machine learning',
  long_description=open('README.md').read(),
  long_description_content_type="text/markdown",
  author = 'Ilan Moscovitz',
  author_email = 'ilan.moscovitz@gmail.com',
  url = 'https://github.com/imoscovitz/wittgenstein',
  keywords = ['Classification', 'Decision Rule', 'Machine Learning', 'Explainable Machine Learning', 'Data Science'],
  packages=setuptools.find_packages(),
  install_requires=[
          'pandas',
          'numpy'
      ],
  classifiers=[
    'Programming Language :: Python',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7'
  ],
  include_package_data=True,
  package_data={'': ['data/*.csv']}
)
